from bedrock_agentcore.runtime import BedrockAgentCoreApp
from llama_agents.client import HandlerData
from llama_agents.server._runtime.idle_release_runtime import IdleReleaseDecorator
from llama_agents.server._runtime.persistence_runtime import PersistenceDecorator
from llama_agents.server._service import _WorkflowService as WorkflowService
from llama_agents.server._store.abstract_workflow_store import AbstractWorkflowStore
from workflows import Workflow
from workflows.events import StartEvent
from workflows.plugins.basic import basic_runtime
from workflows.runtime.types.plugin import Runtime
from workflows.utils import _nanoid

from ._runtime_decorator import AgentCoreRuntimeDecorator

IDLE_TIMEOUT = 60.0
PERSISTENCE_BACKOFF = [0.5, 3]
POLLING_TIMEOUT = 600


class WorkflowNotFoundError(Exception):
    """
    Raise when a workflow with a given name cannot be found within the registered workflows.
    """

    def __init__(self, workflow_name: str) -> None:
        self.workflow_name = workflow_name

    def __repr__(self) -> str:
        return f"Could not find {self.workflow_name} among the registered workflows"

    def __str__(self) -> str:
        return f"Could not find {self.workflow_name} among the registered workflows"


class AgentCoreService:
    def __init__(
        self,
        app: BedrockAgentCoreApp,
        store: AbstractWorkflowStore,
        idle_timeout: float = IDLE_TIMEOUT,
        persistence_backoff: list[float] = PERSISTENCE_BACKOFF,
    ) -> None:
        self._workflow_store = store
        inner: Runtime = IdleReleaseDecorator(
            PersistenceDecorator(basic_runtime, store=self._workflow_store),
            store=self._workflow_store,
            idle_timeout=idle_timeout,
        )
        self._runtime: AgentCoreRuntimeDecorator = AgentCoreRuntimeDecorator(
            decorated=inner,
            store=self._workflow_store,
            app=app,
            persistence_backoff=persistence_backoff,
        )
        self._service = WorkflowService(
            runtime=self._runtime, store=self._workflow_store
        )

    def add_workflow(self, workflow_name: str, workflow: Workflow) -> None:
        workflow._switch_workflow_name(workflow_name)
        workflow._switch_runtime(self._runtime)

    async def run_workflow(
        self, workflow_name: str, start_event: StartEvent
    ) -> HandlerData:
        wf = self._service.get_workflow(workflow_name)
        if wf is not None:
            handler_data = await self._service.start_workflow(
                wf, start_event=start_event, handler_id=_nanoid()
            )
            return await self._service.await_workflow(handler_data)

        raise WorkflowNotFoundError(workflow_name)



import asyncio
from typing import Literal, cast

from bedrock_agentcore.memory import MemoryClient

ActivityStatus = Literal["ACTIVE", "FAILED", "CREATING", "DELETING"]


async def check_memory_activity(client: MemoryClient, memory_id: str) -> bool:
    max_retries = 120
    base_delay = 0.5
    retries = 0
    while retries < max_retries:
        status = await asyncio.to_thread(
            client.get_memory_status,
            memory_id=memory_id,
        )
        status = cast(ActivityStatus, status)
        if status in ("FAILED", "DELETING"):
            return False
        elif status == "ACTIVE":
            return True
        else:
            if retries < max_retries - 1:
                delay = base_delay * (retries + 1)
                await asyncio.sleep(delay)
            retries += 1
    return False


async def create_memory(client: MemoryClient, memory_name: str, namespace: str) -> str:
    response = await asyncio.to_thread(
        client.create_or_get_memory,
        name=memory_name,
    )
    memory_id = response.get("memoryId", response.get("id"))
    if memory_id is None:
        raise RuntimeError(f"AgentCore failed to create a memory ID for {namespace}")
    return memory_id



from bedrock_agentcore.runtime import BedrockAgentCoreApp
from llama_agents.server._runtime.server_runtime import ServerRuntimeDecorator
from llama_agents.server._store.abstract_workflow_store import AbstractWorkflowStore
from workflows import Workflow
from workflows.events import Event
from workflows.runtime.types.plugin import (
    RegisteredWorkflow,
    Runtime,
)
from workflows.runtime.types.results import StepFunctionResult, StepWorkerState
from workflows.runtime.types.step_function import (
    StepWorkerFunction,
    as_step_worker_functions,
    create_workflow_run_function,
)


def as_agentcore_async_task(
    app: BedrockAgentCoreApp, name: str, fn: StepWorkerFunction
) -> StepWorkerFunction:
    async def step_worker(
        state: StepWorkerState,
        step_name: str,
        event: Event,
        workflow: Workflow,
    ) -> list[StepFunctionResult]:
        task_id = app.add_async_task(name)
        try:
            results = await fn(state, step_name, event, workflow)
        finally:
            app.complete_async_task(task_id)
        return results

    return step_worker


class AgentCoreRuntimeDecorator(ServerRuntimeDecorator):
    def __init__(
        self,
        decorated: Runtime,
        store: AbstractWorkflowStore,
        app: BedrockAgentCoreApp,
        *,
        persistence_backoff: list[float] | None = None,
    ) -> None:
        super().__init__(decorated, store, persistence_backoff=persistence_backoff)
        self.app = app
        self._tracked: dict[str, Workflow] = {}
        self._registered: dict[int, RegisteredWorkflow] = {}

    def register(self, workflow: Workflow) -> RegisteredWorkflow:
        name = workflow.workflow_name
        if name not in self._tracked:
            self._tracked[name] = workflow
        wrapped_steps: dict[str, StepWorkerFunction] = {
            step_name: as_agentcore_async_task(self.app, f"{name}.{step_name}", step)
            for step_name, step in as_step_worker_functions(workflow).items()
        }
        registered = RegisteredWorkflow(
            steps=wrapped_steps,
            workflow_run_fn=create_workflow_run_function(workflow),
            workflow=workflow,
        )
        id_ = id(workflow)
        self._registered[id_] = registered
        return registered

    def track_workflow(self, workflow: Workflow) -> None:
        self._tracked[workflow.workflow_name] = workflow
        super().track_workflow(workflow)

    def untrack_workflow(self, workflow: Workflow) -> None:
        self._tracked.pop(workflow.workflow_name, None)
        super().untrack_workflow(workflow)

    def get_registered(self, workflow: Workflow) -> RegisteredWorkflow | None:
        return self._registered.get(id(workflow))

    def get_workflow(self, name: str) -> Workflow | None:
        return self._tracked.get(name)

    def get_workflow_names(self) -> list[str]:
        return list(self._tracked.keys())
